/*  File: analyser-EEJets.cc
 *  Created by Zoltán Nagy on 14 January 2020.
 *  Modified by D. Soper on 30 January 2021.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


/*   local headers  */
#include "analyzer-EEJets.h"



void UserEEJets::initfunc()
{
  /*   Creating histograms to calculate the cross sections   */
  Hist1D::phys(1, "Thrust distribution * (1-T)", spacing<hist1d>::logarithmic(100, 0.0001, 0.5), true);
  Hist1D::phys(2, "Thrust distribution * (1-T), alt", spacing<hist1d>::logarithmic(100, 0.0001, 0.5), true);
  Hist1D::phys(3, "Two jet probability", spacing<hist1d>::logarithmic(24, 0.000001, 1.0), true);
  Hist1D::phys(4, "y23 distribution", spacing<hist1d>::logarithmic(48, 0.000001, 1.0), true);
}


void UserEEJets::userfunc(const shower_stage<epa>& hard,
                          const shower_stage<epa>& /*  afterthreshold  */,
                          const shower_stage<epa>& showered)
{
  /*   weights   */
  double wh = weight(hard);
  double ws = weight(showered);
    
  /*    Fill the thrust histograms   */
  
  double T = thrust(showered.state);
  Hist1D::physfill(1, (1.0-T)*ws, dirac{}, 1.0-T);
  Hist1D::normfill(1, wh);
  Hist1D::physfill(2, (1.0-T)*ws, dirac{}, 1.0-T);
  Hist1D::normfill(2, ws);
  
  /*    Fill the y23 histograms   */
  double y23value = y23(showered.state);
  Hist1D::physfill(4, y23value*ws, dirac{}, y23value);
  Hist1D::normfill(4, ws);
  
  /*  Fill the two jet histogram   */

  Hist1D::normfill(3, ws);
  for(int n=0;n<24;++n)
  {
    double ycut = 0.00000133352*pow(10.0,n/4.0); // 10.0^(1/8) = 1.33352
    int N = Njets(showered.state, ycut);
    //cout << ycut << " Njets = " << N << endl;
    
    if(N<3)
    {
      double binwidth = pow(10.0,(n+1)/4.0-6.0)-pow(10.0,n/4.0-6.0);
      Hist1D::physfill(3, ws*binwidth, dirac{}, ycut);
    }
  }
}

double UserEEJets::y23(const event<epa>& pIN)
{
  energy<double> E = mag(pIN[-1].momentum + pIN[0].momentum);
  auto[lo,np] = bounds_of<colored,fin>(pIN);
  /* Note that y23 here is dimensionless. */
  double y23 = 0.0;
  
  /* Define the parton momentum vectors */
  std::vector<duct::lorentzvector<double>> p;
  for(int i = 1; i <= np; ++i)
  {
    p.push_back(pIN[i].momentum/E);
  }
  if(np == 2)
    {
      y23 = 0.0;
      return y23;
    }
  /* We loop over vij steps.
     Each step reduces np by one and possibly increases Njets by 1. */
  while(np > 2)
  {
    /* vij = 2*(1 - cos(theta_ij))
       so vij cannot be bigger than 4. */
    double vijmin = 100.0;
    int imin = 0;
    int jmin = 0;
    /* Loop over partons i and j with i < j
       to find the minimum value of vij and the corresponding i and j. */
    for(int i = 0; i < np-1; ++i)
      for(int j = i + 1; j < np; ++j)
      {
        double vij = 2.0*(1.0 - unit(p[i])*unit(p[j]));
        if (vij < vijmin)
        {
          vijmin = vij;
          imin = i;
          jmin = j;
        }
      }
    /* Let imin label the parton with smaller energy */
    if(p[jmin][0] < p[imin][0]) std::swap(imin,jmin);
    /* Compute yij for the chosen pair of partons */
    double Eimin = p[imin][0];
    double yij = vijmin*Eimin*Eimin;
    
    /* We will find the largest yij produced. */
    if(yij > y23) y23 = yij;

    lorentzvector<double> pij = p[imin] + p[jmin];
    p[jmin] = pij;
    p.erase(p.begin() + imin);
    np--;
    if(np==2) return y23;
  }
  cout << "We should not get to here. y23 = "<< y23 << endl;
  return y23;
}

int UserEEJets::Njets(const event<epa>& pIN, const double& ycut)
{
  energy<double> E = mag(pIN[-1].momentum + pIN[0].momentum);
  /* Note that ycut here is in GeV^2 units, as in the original paper. */
  int Njets = 0;
  auto[lo,np] = bounds_of<colored,fin>(pIN);
  /* Define the parton momentum vectors */
  std::vector<duct::lorentzvector<double>> p;
  for(int i = 1; i <= np; ++i)
  {
    p.push_back(pIN[i].momentum/E);
  }

  /* We loop over vij steps.
     Each step reduces np by one and possibly increases Njets by 1. */
  while(np > 1)
  {
    /* vij = 2*(1 - cos(theta_ij))
       so vij cannot be bigger than 4. */
    double vijmin = 100.0;
    int imin = 0;
    int jmin = 0;
    /* Loop over partons i and j with i < j
       to find the minimum value of vij and the corresponding i and j. */
    for(int i = 0; i < np-1; ++i)
      for(int j = i + 1; j < np; ++j)
      {
        double vij = 2.0*(1.0 - unit(p[i])*unit(p[j]));
        if (vij < vijmin)
        {
          vijmin = vij;
          imin = i;
          jmin = j;
        }
      }
    /* Let imin label the parton with smaller energy */
    if(p[jmin][0] < p[imin][0]) std::swap(imin,jmin);
    /* Compute yij for the chosen pair of partons */
    double Eimin = p[imin][0];
    double yij = vijmin*Eimin*Eimin;
    if(yij < ycut)
    {
      lorentzvector<double> pij = p[imin] + p[jmin];
      p[jmin] = pij;
    }
    else
    {
      Njets++;
    }
    p.erase(p.begin() + imin);
    np--;
    if(Njets>1) // We return 3 for Njets =>3
    {
      Njets++;
      return Njets;
    }
    if(np==1)
    {
      Njets++;
      return Njets;
    }
  }
  cout << "We should not get to here. Njets = "<< Njets << endl;
  return Njets;
}

double UserEEJets::thrust(const event<epa>& p)
{
  energy<double> E = mag(p[-1].momentum + p[0].momentum);
  double thrust = 0.0;
  auto[lo,np] = bounds_of<colored,fin>(p);
  
  /*    Loop over reference vectors orthogonal to two particles.    */
  for(int i = 1; i < np; ++i)
    for(int j = i + 1; j <= np; ++j)
    {
      auto part = threevector<double>{0.0,0.0,0.0};
      auto axis = cross(p[i].momentum, p[j].momentum)/E/E;
      
      for(int k = 1; k <= np; ++k)
        if(k != i && k != j) {
          if(dot(p[k].momentum/E, axis) > 0.0) part += p[k].momentum/E;
          else part -= p[k].momentum/E;
        }
      
      /*    There are now four cases to be examined.   */
      if(double trial = mag(part + p[i].momentum/E + p[j].momentum/E);
         trial > thrust) thrust = trial;
      if(double trial = mag(part + p[i].momentum/E - p[j].momentum/E);
         trial > thrust) thrust = trial;
      if(double trial = mag(part - p[i].momentum/E + p[j].momentum/E);
         trial > thrust) thrust = trial;
      if(double trial = mag(part - p[i].momentum/E - p[j].momentum/E);
         trial > thrust) thrust = trial;
    }
  
  /*    return the normalized thrust    */
  return thrust;
}
  
